/*
 * (c) Copyright 2001-2003
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.0 - Aug 25 2003 11:11:11
 */

#include <windows.h>
#include <commdlg.h>
#include <commctrl.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <mmsystem.h>

#include "resource.h"
#include "enviromt.h"
#include "ansiterm.h"
#include "ascii.h"

#include "s68.h"
#include "conf.h"
#include "decoder.h"
#include "m68000.h"

#include "s68stat.h"


HINSTANCE  hInstance;         /* Handle to main instance data*/

Conf       conf;
int        quiet = 0;


static HACCEL hAccel;         /* Handle to accelerator table */
static HWND   ws68Wnd;        /* Handle to main window */
static HWND   toolbarWnd;     /* Handle to toolbar window */
static HWND   statusWnd;      /* Handle to status window */
static HWND   textWnd;        /* Hanlde to text window */

static ATTERM *tty[8];

static char   *ws68WinClass = "ws68WClass";
static char   *textWndClass = "textWClass";

static char   cfgFileName[MAX_PATH];    /* .cfg file name */
static int    simStart  = 0;
static int    quietMode = 0;

#define MAX_TEXT_LINES 80

static LPSTR      cText[MAX_TEXT_LINES];
static TEXTMETRIC tmTextInfo;
static int   iMaxNumLines;       // Number of lines that will fit in the TextWnd display.
static int   iCurLine;           // Line # in the text display we're on.
static int   iTextLines;         // # of lines of text which are valid in cText.
static RECT  rText;
static HFONT textFont;

static TBBUTTON tbButtons [] = {
  { 0, ID_FILE_NEW,      TBSTATE_ENABLED,       TBSTYLE_BUTTON, 0L, 0},
  { 1, ID_FILE_OPEN,     TBSTATE_ENABLED,       TBSTYLE_BUTTON, 0L, 0},
  { 2, ID_FILE_SAVE,     TBSTATE_INDETERMINATE, TBSTYLE_BUTTON, 0L, 0},
  { 0, 0,                TBSTATE_ENABLED,       TBSTYLE_SEP,    0L, 0},
  { 3, ID_EDIT_CUT,      TBSTATE_INDETERMINATE, TBSTYLE_BUTTON, 0L, 0},
  { 4, ID_EDIT_COPY,     TBSTATE_INDETERMINATE, TBSTYLE_BUTTON, 0L, 0},
  { 5, ID_EDIT_PASTE,    TBSTATE_INDETERMINATE, TBSTYLE_BUTTON, 0L, 0},
  { 0, 0,                TBSTATE_ENABLED,       TBSTYLE_SEP,    0L, 0},
  { 6, ID_SIM_START,     TBSTATE_ENABLED,       TBSTYLE_BUTTON, 0L, 0},
  { 7, ID_SIM_STOP,      TBSTATE_INDETERMINATE, TBSTYLE_BUTTON, 0L, 0},
  { 8, ID_SIM_RESET,     TBSTATE_ENABLED,       TBSTYLE_BUTTON, 0L, 0},
  { 9, ID_SIM_TRACE,     TBSTATE_ENABLED,       TBSTYLE_BUTTON, 0L, 0},
  { 0, 0,                TBSTATE_ENABLED,       TBSTYLE_SEP,    0L, 0},
  {10, ID_HELP_CONTENTS, TBSTATE_ENABLED,       TBSTYLE_BUTTON, 0L, 0}
};

#define TB_ENTRIES sizeof(tbButtons)/sizeof(tbButtons[0])



/*
 * d i a l o g A b o u t P r o c
 */
BOOL CALLBACK  dialogAboutProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)  {
	  case WM_INITDIALOG:
		return TRUE;

	  case WM_SYSCOMMAND:
		switch (wParam) {
			case SC_CLOSE:
				EndDialog(hWnd, 0);
				return TRUE;
			default:
				break;
			}
			break;

	  case WM_COMMAND:
		switch (LOWORD(wParam)){
		  case IDCANCEL:
		  case IDOK:
			EndDialog(hWnd, 1);
			return TRUE;
		  default:
			break;
		}
		break;
	  default:
		break;
	}

	return FALSE;
}



/*
 * d i a l o g T e r m P r o c
 */
BOOL CALLBACK dialogTermProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	ATTERM *atTerm;
	BOOL   state;
	long   i;

	switch (msg) {
	  case WM_INITDIALOG:
		atTerm = (ATTERM *)lParam;
		if (!atTerm)
			EndDialog(hWnd, IDCANCEL);

		/* Emulation */
		switch (atTerm->aEmulation) {
		  case AT_EMU_VT100:
			CheckRadioButton(hWnd, IDC_ATOPTEMUNONE, IDC_ATOPTEMUVT100, IDC_ATOPTEMUVT100);
			break;
		  case AT_EMU_NONE:
		  default:
			CheckRadioButton(hWnd, IDC_ATOPTEMUNONE, IDC_ATOPTEMUVT100, IDC_ATOPTEMUNONE);
			break;
		}

//		/* Local echo */
//		if (atTerm->aLocalEcho)
//			CheckDlgButton(hWnd, IDC_ATOPTLOCALECHO, TRUE);
//		else
//			CheckDlgButton(hWnd, IDC_ATOPTLOCALECHO, FALSE);

		/* Bell */
		if (atTerm->aBell)
			CheckDlgButton(hWnd, IDC_ATOPTBELL, TRUE);
		else
			CheckDlgButton(hWnd, IDC_ATOPTBELL, FALSE);

		/* Backspace Character */
		if (atTerm->aBackSpaceChar == ASCII_BS)
			CheckRadioButton(hWnd, IDC_ATOPTBSBS, IDC_ATOPTBSDEL, IDC_ATOPTBSBS);
		else
			CheckRadioButton(hWnd, IDC_ATOPTBSBS, IDC_ATOPTBSDEL, IDC_ATOPTBSDEL);

		/* Delete Character */
		if (atTerm->aDeleteChar == ASCII_BS)
			CheckRadioButton(hWnd, IDC_ATOPTDELBS, IDC_ATOPTDELDEL, IDC_ATOPTDELBS);
		else
			CheckRadioButton(hWnd, IDC_ATOPTDELBS, IDC_ATOPTDELDEL, IDC_ATOPTDELDEL);

//		/* Scrollback Buffer */
//		SetDlgItemInt(hWnd, IDC_ATOPTSCROLLBUF,
//			atTerm->aTermBuf->tBufY-atTerm->aTermBuf->tScrY-ATIB_TERMBUF_EXTRALINES,FALSE);
//		return TRUE;

	  case WM_SYSCOMMAND:
		/* Process Control Box / Max-Min function */
		switch (wParam) {
		  case SC_CLOSE:
			EndDialog(hWnd, IDCANCEL);
			return TRUE;
		  default:
			break;
		}
		break;

	  case WM_COMMAND:
		switch (LOWORD(wParam))  {
		  case IDOK:
			atTerm = tty[0];

			/* Emulation */
			if (IsDlgButtonChecked(hWnd, IDC_ATOPTEMUVT100)) {
				atTerm->aEmulation = AT_EMU_VT100;
			} else {
				atTerm->aEmulation = AT_EMU_NONE;
			}

//			/* Local echo */
//			atTerm->aLocalEcho = IsDlgButtonChecked(hWnd, IDC_ATOPTLOCALECHO);
//			atTerm->aVLocalEcho = atTerm->aLocalEcho;
			/* Bell */
			atTerm->aBell = IsDlgButtonChecked(hWnd, IDC_ATOPTBELL);

			/* Backspace Character */
			if (IsDlgButtonChecked(hWnd, IDC_ATOPTBSBS)) {
				atTerm->aBackSpaceChar = ASCII_BS;
			} else {
				atTerm->aBackSpaceChar = ASCII_DEL;
			}

			/* Delete Character */
			if (IsDlgButtonChecked(hWnd, IDC_ATOPTDELBS)) {
				atTerm->aDeleteChar = ASCII_BS;
			} else {
				atTerm->aDeleteChar = ASCII_DEL;
			}

			/* Scrollback buffer */
			i = GetDlgItemInt(hWnd, IDC_ATOPTSCROLLBUF, &state, FALSE);
			if (state) {
				atAdjustTerm(atTerm, i);
			}

			EndDialog(hWnd, IDOK);
			return TRUE;

		  case IDCANCEL:
			EndDialog(hWnd, IDCANCEL);
			return TRUE;

		  default:
			break;
		}
		break;

	  case WM_DESTROY:
		break;

	  default:
		break;
	}

	return FALSE;
}



/*
 * d i a l o g F n K e y P r o c
 *
 */
BOOL CALLBACK  dialogFnKeyProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	ATTERM *atTerm;
	int    i, j, k;

	switch (msg)  {
	  case WM_INITDIALOG:
		atTerm = (ATTERM *)lParam;
		if (!atTerm)
			EndDialog(hWnd, IDCANCEL);
		CheckRadioButton(hWnd, IDC_ATFNKEYBANKNONE, IDC_ATFNKEYBANKSHCT, IDC_ATFNKEYBANKNONE);
		for (i = 0; i < 12; i++) {
			SetDlgItemText(hWnd, IDC_ATFNKEYFN1+i,(char *)atTerm->aFnKeyMacro[i]);
		}
		return TRUE;

	  case WM_SYSCOMMAND:
		/* Process Control Box / Max-Min function */
		switch(wParam) {
			case SC_CLOSE:
				EndDialog(hWnd, IDCANCEL);
				return TRUE;
			default:
				break;
		}
		break;

	  case WM_COMMAND:
		switch (LOWORD(wParam)) {
		  case IDOK:
		  case IDC_ATFNKEYBANKNONE:
		  case IDC_ATFNKEYBANKSHIFT:
		  case IDC_ATFNKEYBANKCTRL:
		  case IDC_ATFNKEYBANKSHCT:
			atTerm = tty[0];
			j = ((int)(GetProp(hWnd, "FnPage"))) * 12;
			if(IsDlgButtonChecked(hWnd, IDC_ATFNKEYBANKNONE)) {
				k = 0;
//				SetProp(hWnd, "FnPage", (HANDLE)0);
			} else if(IsDlgButtonChecked(hWnd, IDC_ATFNKEYBANKSHIFT)) {
				k = 12;
//				SetProp(hWnd, "FnPage", (HANDLE)1);
			} else if(IsDlgButtonChecked(hWnd, IDC_ATFNKEYBANKCTRL)) {
				k = 24;
//				SetProp(hWnd, "FnPage", (HANDLE)2);
			} else {
				k = 36;
//				SetProp(hWnd, "FnPage", (HANDLE)3);
			}
			for (i = 0; i < 12; i++) {
				/* first save our old setting */
				GetDlgItemText(hWnd, IDC_ATFNKEYFN1 + i,
					(char*)atTerm->aFnKeyMacro[i+j], TBMACRO_MAXLEN);
				/* force a terminating 0x00 - I don't trust microshit */
				atTerm->aFnKeyMacro[i+j][TBMACRO_MAXLEN-1] = 0x00;
				/* now set our new setting */
					SetDlgItemText(hWnd, IDC_ATFNKEYFN1+i, (char*)atTerm->aFnKeyMacro[i+k]);
			}
			if (LOWORD(wParam) == IDOK) {
				/* update our options */
				atTerm->aBell = atTerm->aBell;
				EndDialog(hWnd, IDOK);
			}
			return TRUE;

		  case IDCANCEL:
			EndDialog(hWnd, IDCANCEL);
			return TRUE;

		  default:
			break;
		}
		break;

	  case WM_DESTROY:
		break;

	  default:
		break;
	}

	return FALSE;
}



/*
 * t o o l b a r C r e a t e W n d
 *
 * Create the toolbar window
 */
BOOL toolbarCreateWnd(HINSTANCE instance, HWND parent)
{
	toolbarWnd = CreateToolbarEx(parent,
        WS_CHILD | WS_BORDER | WS_VISIBLE | TBSTYLE_TOOLTIPS | CCS_ADJUSTABLE, // window style
        IDR_TOOLBAR,  // toolbar ID
        8,            // number of bitmaps
        instance,     // mod instance
        IDB_TOOLBAR,  // resource ID for bitmap
        (LPCTBBUTTON)&tbButtons, // address of buttons
        TB_ENTRIES,   // number of buttons
        16, 16,       // width & height of buttons
        16, 16,       // width & height of bitmaps
        sizeof (TBBUTTON)); // structure size

    if (!toolbarWnd)
        return FALSE;

    return TRUE;
}



/*
 * t e x t R e f r e s h
 */
void textRefresh(void)
{
    HDC  hDC;
    int  i;

    hDC = GetDC(textWnd);
    iCurLine = 0;

	SelectObject(hDC, textFont);
    for(i = 0; i < iMaxNumLines; i++) {
        if ((iTextLines - (iMaxNumLines-i)) >= 0) {
            TextOut(hDC, 0, iCurLine * tmTextInfo.tmHeight + 2,
                    cText[iTextLines-(iMaxNumLines-i)],
                    lstrlen(cText[iTextLines-(iMaxNumLines-i)]));
            iCurLine++;
        }
     }
     ReleaseDC(textWnd, hDC);
}


LPSTR lstrdup(LPSTR s)
{
    HGLOBAL hMem;
    LPSTR   lpRet;

    hMem = GlobalAlloc(GHND, lstrlen(s)+1);
    lpRet = GlobalLock(hMem);
    lstrcpy(lpRet,s);

    return(lpRet);
}



/*
 * t e x t P r i n t L i n e
 */
void textPrintLine(LPSTR cLine)
{
    HDC     hDC;
    int     i;
    RECT    rScroll;
    HGLOBAL hMem;

    hDC = GetDC(textWnd);

    //
    // If we already have MAX_TEXT_LINES of text remembered,
    // we need to zap one to make room for the new line.
    //
    if (iTextLines == MAX_TEXT_LINES) {
        hMem = (HGLOBAL)cText[0];
        GlobalUnlock(hMem);
        GlobalFree(hMem);
        for(i = 0; i < MAX_TEXT_LINES-1; i++) {
            cText[i] = cText[i+1];
        }
        iTextLines--;
    }

    //
    // Copy the line to be displayed into our buffer
    //
    cText[iTextLines++] = lstrdup(cLine);

    //
    // Scroll if necessary
    //
    if (iCurLine >= iMaxNumLines) {
        GetClientRect(textWnd, &rScroll);
        rScroll.top += 2;
        ScrollWindow(textWnd, 0, -tmTextInfo.tmHeight, &rScroll, &rScroll);
        UpdateWindow(textWnd);
        iCurLine--;
    }
    SelectObject(hDC, textFont);
    TextOut(hDC, 0, iCurLine * tmTextInfo.tmHeight + 2,
            cText[iTextLines-1],
            lstrlen(cText[iTextLines-1]));

    iCurLine++;

    ReleaseDC(textWnd, hDC);

    return;
}



/*
 * t e x t W n d P r o c
 */
long CALLBACK textWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT   ps;
    HDC           hDC;
    int           i;

	switch (msg)  {
      case WM_CREATE:
        for (i = 0; i < MAX_TEXT_LINES; i++) {
            cText[i] = (LPSTR) NULL;
        }
        iTextLines = iCurLine = 0;
        hDC = GetDC(hWnd);
        if (!GetTextMetrics(hDC, &tmTextInfo)) {
             MessageBox((HWND)NULL, "Textwindow creat failed",
                         "TextWndProc", MB_ICONEXCLAMATION|MB_APPLMODAL);
        }
        ReleaseDC(hWnd, hDC);
        GetClientRect(hWnd, &rText);
        iMaxNumLines = (rText.bottom - rText.top-5)/tmTextInfo.tmHeight;
        break;

      case WM_PAINT:
        GetClientRect(hWnd, &rText);
        hDC = BeginPaint(hWnd, &ps);
        SetBkMode(hDC, OPAQUE);
        textRefresh();
        EndPaint(hWnd, &ps);
        break;

	  case WM_SIZE:
        GetClientRect(hWnd, &rText);
        iMaxNumLines = (rText.bottom - rText.top - 5)/tmTextInfo.tmHeight;
        textRefresh();

      default:
		 return DefWindowProc(hWnd, msg, wParam, lParam);
	}

	return(0); 
}



/*
 * t e x t I n i t C l a s s
 */
BOOL textInitClass(HINSTANCE hInst)
{
	WNDCLASS  wc;

	wc.style = CS_HREDRAW|CS_VREDRAW;
	wc.lpfnWndProc = textWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 4;
	wc.hInstance = hInst;
	wc.hIcon = (HICON) NULL;
	wc.hCursor = LoadCursor((HINSTANCE)NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = textWndClass;

	return RegisterClass(&wc);
}



/*
 * t e x t C r e a t e W n d
 *
 * Create the text window
 */
BOOL textCreateWnd(HINSTANCE instance, HWND parent)
{
	RECT   rect;
    HDC    hDC;

    if (parent == NULL)  {
	    textWnd = CreateWindow(textWndClass, NULL,
                WS_OVERLAPPEDWINDOW | WS_VSCROLL | WS_VISIBLE,
                0, 0, 640, 480,
                NULL, (HMENU)NULL, instance, NULL);
    }
    else  {
        GetClientRect(parent, &rect);
        textWnd = CreateWindow(textWndClass, NULL,
                WS_CHILD | WS_DLGFRAME | WS_VISIBLE,
                0, 28,
				(rect.right - rect.left), (rect.bottom - rect.top) - 28 - 20,
                parent, (HMENU)NULL, instance, NULL);
    }
    if (!textWnd)
		return FALSE;

    GetClientRect(textWnd, &rText);

	hDC = GetDC(textWnd);
	textFont = GetStockObject(SYSTEM_FIXED_FONT);
	SelectObject(hDC, textFont);
    ReleaseDC(textWnd, hDC);

	return TRUE;
}



/*
 * t e x t D e s t r o y W n d
 */
void textDestroyWnd(void)
{
    int  i;
    HGLOBAL hMem;

    for (i = 0; i < iTextLines ; i++) {
        if (cText[i] != NULL) {
            hMem = (HGLOBAL) cText[i];
            GlobalUnlock(hMem);
            GlobalFree(hMem);
        }
    }
}



/*
 * s t a t u s C r e a t e W n d
 *
 * Create the status window
 */
BOOL statusCreateWnd(HINSTANCE instance, HWND parent)
{
    statusWnd = CreateStatusWindow( 
      WS_CHILD | WS_BORDER | WS_VISIBLE,  // window styles
      "Ready",                            // default window text
      parent,                             // parent window
      IDR_STATUSWND);                     // ID

	if (!statusWnd)
        return FALSE;

    return TRUE;
}


/*
 * r e a d C o n f
 *
 * Read new configuration file
 */
int readConf(char *sFile)
{
    FILE  *sf;
    char  line[1024];
    char  *p;
    int   lineNum = 1;
    char  *pname, *pvalue;
    Dev   *dp;
    Cmd   *cp;

    if ((sf = fopen(sFile, "r")) == NULL)  return(FALSE);

    conf.numDev = 0;
    conf.numCmd = 0;
    conf.rstBase = 0x000000;
    conf.numInstCycle = 1000;

    while ((p = fgets(line, sizeof(line), sf)) != NULL)  {
        pname = strtok(p, " \t\n");
        if ((pname == NULL) || (pname[0] == '\0') || (pname[0] == '#'))
            continue;

        if (!strcmp(pname, "COMMAND"))  {
            /* Get pointer to next command pointer */
            cp = &conf.cmd[conf.numCmd++];

            /* Get command */
            pvalue = strtok(NULL, "\n");
            if ((*cp = realloc(*cp, strlen(pvalue)+1)) == NULL)  {
                uiErrPrintf("%s: Out of memory at line %d\n", sFile, lineNum);
                return(FALSE);
            }
            strcpy(*cp, pvalue);
        }
        else if (!strcmp(pname, "DEVICE"))  {
            /* Get pointer to next device structure */
            dp = &conf.dev[conf.numDev++];

            /* Get device type */
            pvalue = strtok(NULL, " \t\n");
            if ((dp->type = realloc(dp->type, strlen(pvalue)+1)) == NULL)  {
                uiErrPrintf("%s: Out of memory at line %d\n", sFile, lineNum);
                return(FALSE);
            }
            strcpy(dp->type, pvalue);
 
            /* Get base address */
            pvalue = strtok(NULL, " \t\n");
            sscanf(pvalue, "%lx", &dp->base);

            /* Get size */
            pvalue = strtok(NULL, " \t\n");
            sscanf(pvalue, "%lx", &dp->size);

            /* Get interrupt level */
            pvalue = strtok(NULL, " \t\n");
            if (pvalue == NULL)  {
                dp->intlvl = 0;
            }
            else  {
                if (sscanf(pvalue, "%d", &dp->intlvl) != 1)  {
                    uiErrPrintf("%s: Illegal interrupt level at line %d\n", sFile, lineNum);
                    return(FALSE);
                }
            }

            /* Get device arguments */
            pvalue = strtok(NULL, " \t\n");
            if (pvalue == NULL)  {
                dp->devarg = NULL;
            }
            else  {
                if ((dp->devarg = malloc(strlen(pvalue)+1)) == NULL)  {
                    uiErrPrintf("%s: Out of memory at line %d\n", sFile, lineNum);
                    return(FALSE);
                }
                strcpy(dp->devarg, pvalue);
            }
        }
        else if (!strcmp(pname, "CYCLE"))  {
            /* Get number of Inst. per Polling cycle */
            pvalue = strtok(NULL, " \t\n");
            sscanf(pvalue, "%d", &conf.numInstCycle);
        }
        else if (!strcmp(pname, "RSTBASE"))  {
            /* Get base address */
            pvalue = strtok(NULL, " \t\n");
            sscanf(pvalue, "%lx", &conf.rstBase);
        }
        else {
            uiErrPrintf("%s: '%s' unknown keyword at line %d\n", sFile, pname, lineNum);
            return FALSE;
        }
        lineNum++;
    }

    return TRUE;
}


/*
 * m a i n E x i t
 */
void mainExit()
{

    if (!quietMode)  {
        EnableWindow(ws68Wnd, FALSE);

        /* Shut down various modules */
        WinHelp(ws68Wnd, NULL, HELP_QUIT, 0L);
        vt100DestroyWnd(tty[0]);
        textDestroyWnd();
        DestroyWindow(ws68Wnd);
    }
    PostQuitMessage(0);        /* Signal program to end */
}



/*
 * m a i n S e t W i n d o w S i z e
 */
static void mainSetWindowSize(UINT type, int cx, int cy)
{
    if (type == SIZE_MINIMIZED)
        return;

    SendMessage(statusWnd, WM_SIZE, 0, 0);
    SendMessage(toolbarWnd, WM_SIZE, 0, 0);
    SendMessage(textWnd, WM_SIZE, 0, 0);
}



/*
 * m a i n W n d P r o c
 */
long CALLBACK mainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    LPTOOLTIPTEXT lpToolTipText;
    static CHAR   szBuf[128];

	switch (msg) {
	  case WM_NCDESTROY:
		if (IsWindow(ws68Wnd)) {
			PostMessage(ws68Wnd, WM_SYSCOMMAND, SC_CLOSE, 0L);
		}
		break;

	  case WM_DESTROY:
		enSaveWindowPlacement("main", ws68Wnd);
        break;

	  case WM_SYSCOMMAND:
		switch(wParam) {
		  case SC_CLOSE:
			mainExit();
			return(0L);
		  default:
			break;
		}
		break;

      case WM_LBUTTONDBLCLK:
        // Double-click on the toolbar -- display the customization dialog.
        SendMessage(toolbarWnd, TB_CUSTOMIZE, 0L , 0L );
        break;

      case WM_NOTIFY:
        switch (((LPNMHDR) lParam)->code) {
          case TTN_NEEDTEXT:    
            // Display tool tip text.
			lpToolTipText = (LPTOOLTIPTEXT)lParam;
    		LoadString (hInstance, 
	    			lpToolTipText->hdr.idFrom,    // string ID == cmd ID
		    		szBuf,
			    	sizeof(szBuf));
				    lpToolTipText->lpszText = szBuf;
			break;

          case TBN_QUERYDELETE:
            // Toolbar customization -- can we delete this button?
            return TRUE;

          case TBN_GETBUTTONINFO:
            // The toolbar needs information about a button.
            return FALSE;

          case TBN_QUERYINSERT:
            // Can this button be inserted? Just say yo.
            return TRUE;

          case TBN_CUSTHELP:
            // Need to display custom help.
            MessageBox(hWnd, "This help is custom.",NULL, MB_OK);
            break;

          case TBN_TOOLBARCHANGE:
            // Done dragging a bitmap to the toolbar.
            SendMessage(toolbarWnd, TB_AUTOSIZE, 0L, 0L);
            break;

          default:
            return TRUE;
        }
		return 0L;

	  case WM_COMMAND:
		/* Process menu selection event */
		switch (LOWORD(wParam)) {
		  case ID_FILE_OPEN :  {
            OPENFILENAME   ofn;                    // common dialog box structure
            TCHAR  szFile[260] = TEXT("\0");       // buffer for filename
			TCHAR  *szFilter   = TEXT("(Config Files)*.CFG\0*.CFG\0All Files(*.*)\0*.*\0\0"); 

            // Initialize OPENFILENAME
            ofn.lStructSize = sizeof(OPENFILENAME);
            ofn.hwndOwner = ws68Wnd;
            ofn.hInstance = hInstance;
            ofn.lpstrFilter = szFilter;
            ofn.lpstrCustomFilter = (LPTSTR) NULL;
            ofn.nMaxCustFilter = 0;
            ofn.nFilterIndex = 1;
            ofn.lpstrFile = szFile;
            ofn.nMaxFile = sizeof(szFile);
            ofn.lpstrFileTitle = (LPTSTR) NULL;
            ofn.nMaxFileTitle = 0;
            ofn.lpstrInitialDir = (LPSTR) NULL;
            ofn.lpstrTitle = "Open";
            ofn.Flags = OFN_FILEMUSTEXIST;
            ofn.nFileOffset = 0;
            ofn.nFileExtension = 0;
            ofn.lpstrDefExt = "cfg";
            ofn.lCustData = 0;
            ofn.lpfnHook = (LPOFNHOOKPROC) NULL;
            ofn.lpTemplateName = (LPTSTR) NULL;  

            // Display the Open dialog box. 
            if (GetOpenFileName(&ofn) == TRUE)  {
                 int  stat = 0;
				 static char  tmpStr[256];

				 strcpy(cfgFileName, ofn.lpstrFile);
				 stat = readConf(cfgFileName);
                 decoderInit();
                 cpuInit();
                 decoderReset();
                 cpuReset();
				 if (stat == 0)  {
                     wsprintf(tmpStr, "WS68 - %s", cfgFileName);
                     SetWindowText(ws68Wnd, tmpStr);
                     SendMessage(statusWnd, WM_SETTEXT, 0, (LPARAM)"Ready");
				 }
				 else {
                     SendMessage(statusWnd, WM_SETTEXT, 0, (LPARAM)"Failed");
                 }

            }

            } return(0L);
 

		  case ID_FILE_EXIT :
			mainExit();
			return(0L);

		  case ID_OPTIONS_TERMSET :
			if (DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_TERM), ws68Wnd, dialogTermProc, (LPARAM)tty[0]) == IDOK)
                vt100SaveSettings(tty[0], TBNOTIFY_OPTIONS, NULL);
			return 0L;

		  case ID_OPTIONS_FNKEYS : {
			int  i;
			unsigned char macro[48][TBMACRO_MAXLEN];

			for (i = 0; i < 48; i++)
				strcpy((char*)macro[i], (char*)tty[0]->aFnKeyMacro[i]);
			if (DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_FNKEY),
									ws68Wnd, dialogFnKeyProc, (LPARAM)tty[0]) == IDOK) {
                vt100SaveSettings(tty[0], TBNOTIFY_MACROS, NULL);
			} else {
				/* restore macros */
				for (i = 0 ; i < 48; i++)
					strcpy((char*)tty[0]->aFnKeyMacro[i], (char*)macro[i]);
			}
			} return 0L;

		  case ID_OPTIONS_FONT: {
			CHOOSEFONT font;
			LOGFONT    logfont;

			font.lStructSize = sizeof(CHOOSEFONT);
			font.hwndOwner = ws68Wnd;
			font.hDC = NULL;
			font.lpLogFont = &logfont;
			if (tty[0]->aLogFontValid) {
				memcpy(&logfont, &(tty[0]->aLogFont), sizeof(LOGFONT));
				font.Flags = CF_FIXEDPITCHONLY|CF_INITTOLOGFONTSTRUCT|CF_SCREENFONTS;
			} else {
				font.Flags = CF_FIXEDPITCHONLY|CF_SCREENFONTS;
			}
			font.lpTemplateName = NULL;
			font.hInstance = NULL;
			font.lpszStyle = NULL;
			font.nFontType = 0;
			font.nSizeMin = 0;
			font.nSizeMax = 0;
			if (ChooseFont(&font)) {
				vt100ChangeFont(tty[0], NULL, &logfont);
				if (tty[0]->aNotify)  (*(tty[0]->aNotify))(tty[0], TBNOTIFY_FONT);
				vt100SaveSettings(tty[0], TBNOTIFY_FONT, NULL);
			}
			} break;

          case ID_SIM_START :
			simStart = TRUE;
			SendMessage(statusWnd, WM_SETTEXT, 0, (LPARAM)"Running");
			SendMessage(toolbarWnd, TB_SETSTATE, ID_SIM_STOP,  TBSTATE_ENABLED);
			SendMessage(toolbarWnd, TB_SETSTATE, ID_SIM_START, TBSTATE_INDETERMINATE);
			SendMessage(toolbarWnd, TB_SETSTATE, ID_SIM_TRACE, TBSTATE_INDETERMINATE);
			tStart();
			return 0L;

          case ID_SIM_STOP :
			simStart = FALSE;
			SendMessage(statusWnd, WM_SETTEXT, 0, (LPARAM)"Pause");
			SendMessage(toolbarWnd, TB_SETSTATE, ID_SIM_STOP,  TBSTATE_INDETERMINATE);
			SendMessage(toolbarWnd, TB_SETSTATE, ID_SIM_START, TBSTATE_ENABLED);
			SendMessage(toolbarWnd, TB_SETSTATE, ID_SIM_TRACE, TBSTATE_ENABLED);
			tStop();
			tReport();
			return 0L;

          case ID_SIM_RESET :
			decoderReset();
			cpuReset();
			SendMessage(statusWnd, WM_SETTEXT, 0, (LPARAM)"System Reset");
            return 0L;

          case ID_SIM_TRACE :
			SendMessage(statusWnd, WM_SETTEXT, 0, (LPARAM)"Trace");
            trace = 1;
            cpuExec();
			trace = 0;
            SendMessage(statusWnd, WM_SETTEXT, 0, (LPARAM)"Ready");
            return 0L;

		  case ID_HELP_CONTENTS :
			WinHelp(ws68Wnd, "WS68.HLP", HELP_INDEX, 0L);
			return(0L);

		  case ID_HELP_HELP :
		    WinHelp(ws68Wnd, "WINHELP.HLP", HELP_INDEX, 0L);
			return(0L);

		  case ID_HELP_ABOUT :
			DialogBox(hInstance, MAKEINTRESOURCE(IDD_ABOUTBOX), ws68Wnd, dialogAboutProc);
			return(0L);

		  default:
			break;
		}
		break;

	  case WM_SIZE:
		mainSetWindowSize(wParam, LOWORD(lParam), HIWORD(lParam));
		return 0;

	  default:
		break;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}



/*
 * w s 6 8 N o t i f y P r o c
 */
void ws68NotifyProc(struct ATTERMtag *atTerm, int msg)
{
	switch (msg)  {
      case ATNOTIFY_DESTROY :
        if (quietMode)  mainExit();
		  break;
      default :
        break;
	}
}



/*
 * i n i t A p p l i c a t i o n
 * 
 * Initialization for application data - only needs to be once, not every
 * instance.
 */
BOOL initApplication()
{
    WNDCLASS  wc;

    if (!vt100InitClass(hInstance))
        return FALSE;

    if (!textInitClass(hInstance))
		return FALSE;

    wc.style = CS_DBLCLKS;
    wc.lpfnWndProc = mainWndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_WS68));
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)GetStockObject(GRAY_BRUSH);
    wc.lpszMenuName =  MAKEINTRESOURCE(IDM_WS68_MENU);
    wc.lpszClassName = ws68WinClass;

    return RegisterClass(&wc);
}



/*
 * i n i t I n s t a n c e
 *
 * Virtually all initialization commands should go here.
 */
BOOL initInstance(LPSTR cmdLine, int cmdShow)
{
    static WINDOWPLACEMENT *wp;

    if (!quietMode)  {
        hAccel = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDA_WS68));

        ws68Wnd = CreateWindow(ws68WinClass, "WS68",
			     WS_OVERLAPPEDWINDOW,
			     CW_USEDEFAULT, CW_USEDEFAULT, 320, 100,
			     NULL, NULL, hInstance, NULL);
        if (!ws68Wnd)
            return FALSE;

        wp = enRestoreWindowPlacement("main", 0, 0, 640, 240);
        wp->showCmd = SW_SHOWNORMAL;
        SetWindowPlacement(ws68Wnd, wp);

        if (!toolbarCreateWnd(hInstance, ws68Wnd))
            return FALSE;

        if (!statusCreateWnd(hInstance, ws68Wnd))
            return FALSE;

        if (!textCreateWnd(hInstance, ws68Wnd))
            return FALSE;

        ShowWindow(ws68Wnd, cmdShow);
        ShowWindow(textWnd, cmdShow);
        UpdateWindow(ws68Wnd);

        SetActiveWindow(ws68Wnd);
	}

    tty[0] = vt100CreateWnd(hInstance, NULL, 80, 50, 80, 25, "tty0");
	tty[0]->aNotify = ws68NotifyProc;
//    vt100LoadSettings(tty[0], TBNOTIFY_OPTIONS, NULL);
//    vt100LoadSettings(tty[0], TBNOTIFY_MACROS, NULL);
//    vt100LoadSettings(tty[0], TBNOTIFY_FONT, NULL);

    return (TRUE);
}



/*
 * u i P r i n t f
 *
 * User interface printf function
 */
void uiPrintf(char *fmt, ...)
{
    static char  lineBuf[512];
    static int   linePos = 0;
    char      buf[512];
    va_list   ap;
    int       i;

    if (!textWnd)  return;

    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    for (i = 0; buf[i]; i++)  {
        if (buf[i] == '\n')  {
			lineBuf[linePos] = '\0';
            textPrintLine(lineBuf);
            linePos = 0;
		}
		else if (buf[i] == '\t')  {
            do {
                lineBuf[linePos++] = ' ';
            } while ((linePos % 8) != 0);
		}
		else  {
			lineBuf[linePos++] = buf[i];
		}
	}
}



/*
 * u i P u t c h a r
 *
 * User interface function to output one charater
 */
void uiPutchar(unsigned char c)
{
	vt100Send(tty[0], &c, 1);
}



/*
 * u i G e t c h a r
 *
 * User interface function to get one charater
 */
int uiGetchar(void)
{
	unsigned char buf[2];

	if (vt100ReadInput(tty[0], buf, 1) == 1)  return((int) buf[0]);

	return(-1);
}



/*
 * u i E r r P r i n t f
 *
 * User Interface Error printf function
 */
void uiErrPrintf(char *fmt, ...)
{
    char      buf[512];
    va_list   ap;

    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);

	MessageBox(NULL, buf, "WS68: Fatal Error.", MB_OK|MB_APPLMODAL|MB_ICONEXCLAMATION);

}



/*
 * W i n M a i n
 *
 * Main program entry
 */
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInstance, LPSTR cmdLine, int cmdShow)
{
    MSG   msg;        /* Current message */
    char  *p, *p2;
    TCHAR  tmpStr[256];

    hInstance = hInst;
    strcpy(cfgFileName, "s68.cfg");

    InitCommonControls();

    p = cmdLine;

    while (*p) {
		/* Move up to first non-space character */
		while (*p == ' ')  p++;
		if (!(*p))
			/* End of line */
			break;
        if (*p == '-')  {
			p++;
			switch (*p++)  {
              case 'q' :
				quietMode = 1;
				break;

			  default :
				wsprintf(tmpStr, "  M68000 System Simulator  Version %s  \n\n  Usage: ws68 [-q] [<cfg-file>]\n", S68VER);
				MessageBox(NULL, tmpStr, "WS68", MB_OK);
				return 0;
			}
		}
		else  {
			/* Get config file name */
			p2 = cfgFileName;
			*p2 = 0;
			while ((*p != ' ') && (*p)) {
				*p2++ = *p++;
			}
			*p2 = 0;
			break;
		}
	}


    if (!hPrevInstance)  {
        if (!initApplication())  return FALSE;
    }
    if (!initInstance(cmdLine, cmdShow))  {
        return FALSE;
    }

    if (!readConf(cfgFileName))  {
		wsprintf(tmpStr, "Can't open configuration file '%s'", cfgFileName);
		MessageBox(NULL, tmpStr, "WS68", MB_OK);
		cfgFileName[0] = '\0';
    }
	else  {
		if (decoderInit() != 0)  return FALSE;
		if (cpuInit() != 0)  return FALSE;
		decoderReset();
		cpuReset();
	}

    if (quietMode)  {
        simStart = 1;
	}
    else  {
        wsprintf(tmpStr, "WS68 - %s", cfgFileName);
        SetWindowText(ws68Wnd, tmpStr);
        SendMessage(statusWnd, WM_SETTEXT, 0, (LPARAM)"Ready");
		SetFocus(ws68Wnd);
	}

    while (1)  {
		while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
            if (msg.message == WM_QUIT)
                return 0;
		    /* Only translate message if it is not an accelerator message */
			if (!TranslateAccelerator(ws68Wnd, hAccel, &msg))	{
				TranslateMessage(&msg);
		        DispatchMessage(&msg);
			}
		}
		if (simStart)  {
			int  cnt;
			for (cnt = 5000; cnt; cnt--)  {
				cpuExec();
				instCounter++;
			}
		}
		else  {
			WaitMessage();
		}
    }


    return (msg.wParam);
}
